from pyinfra.api import OperationError


def _create_container(**kwargs):
    command = []

    networks = kwargs["networks"] if kwargs["networks"] else []
    ports = kwargs["ports"] if kwargs["ports"] else []
    volumes = kwargs["volumes"] if kwargs["volumes"] else []
    env_vars = kwargs["env_vars"] if kwargs["env_vars"] else []

    if kwargs["image"] == "":
        raise OperationError("missing 1 required argument: 'image'")

    command.append("docker container create --name {0}".format(kwargs["container"]))

    for network in networks:
        command.append("--network {0}".format(network))

    for port in ports:
        command.append("-p {0}".format(port))

    for volume in volumes:
        command.append("-v {0}".format(volume))

    for env_var in env_vars:
        command.append("-e {0}".format(env_var))

    if kwargs["pull_always"]:
        command.append("--pull always")

    command.append(kwargs["image"])

    if kwargs["start"]:
        command.append("; {0}".format(_start_container(container=kwargs["container"])))

    return " ".join(command)


def _remove_container(**kwargs):
    return "docker container rm -f {0}".format(kwargs["container"])


def _start_container(**kwargs):
    return "docker container start {0}".format(kwargs["container"])


def _stop_container(**kwargs):
    return "docker container stop {0}".format(kwargs["container"])


def _pull_image(**kwargs):
    return "docker image pull {0}".format(kwargs["image"])


def _remove_image(**kwargs):
    return "docker image rm {0}".format(kwargs["image"])


def _prune_command(**kwargs):
    command = ["docker system prune"]

    if kwargs["all"]:
        command.append("-a")

    if kwargs["filter"] != "":
        command.append("--filter={0}".format(kwargs["filter"]))

    if kwargs["volumes"]:
        command.append("--volumes")

    command.append("-f")

    return " ".join(command)


def _create_volume(**kwargs):
    command = []
    labels = kwargs["labels"] if kwargs["labels"] else []

    command.append("docker volume create {0}".format(kwargs["volume"]))

    if kwargs["driver"] != "":
        command.append("-d {0}".format(kwargs["driver"]))

    for label in labels:
        command.append("--label {0}".format(label))

    return " ".join(command)


def _remove_volume(**kwargs):
    return "docker image rm {0}".format(kwargs["volume"])


def _create_network(**kwargs):
    command = []
    opts = kwargs["opts"] if kwargs["opts"] else []
    ipam_opts = kwargs["ipam_opts"] if kwargs["ipam_opts"] else []
    labels = kwargs["labels"] if kwargs["labels"] else []

    command.append("docker network create {0}".format(kwargs["network"]))
    if kwargs["driver"] != "":
        command.append("-d {0}".format(kwargs["driver"]))

    if kwargs["gateway"] != "":
        command.append("--gateway {0}".format(kwargs["gateway"]))

    if kwargs["ip_range"] != "":
        command.append("--ip-range {0}".format(kwargs["ip_range"]))

    if kwargs["ipam_driver"] != "":
        command.append("--ipam-driver {0}".format(kwargs["ipam_driver"]))

    if kwargs["subnet"] != "":
        command.append("--subnet {0}".format(kwargs["subnet"]))

    if kwargs["scope"] != "":
        command.append("--scope {0}".format(kwargs["scope"]))

    if kwargs["ingress"]:
        command.append("--ingress")

    if kwargs["attachable"]:
        command.append("--attachable")

    for opt in opts:
        command.append("--opt {0}".format(opt))

    for opt in ipam_opts:
        command.append("--ipam-opt {0}".format(opt))

    for label in labels:
        command.append("--label {0}".format(label))
    return " ".join(command)


def _remove_network(**kwargs):
    return "docker network rm {0}".format(kwargs["network"])


def handle_docker(resource, command, **kwargs):
    container_commands = {
        "create": _create_container,
        "remove": _remove_container,
        "start": _start_container,
        "stop": _stop_container,
    }

    image_commands = {
        "pull": _pull_image,
        "remove": _remove_image,
    }

    volume_commands = {
        "create": _create_volume,
        "remove": _remove_volume,
    }

    network_commands = {
        "create": _create_network,
        "remove": _remove_network,
    }

    system_commands = {
        "prune": _prune_command,
    }

    docker_commands = {
        "container": container_commands,
        "image": image_commands,
        "volume": volume_commands,
        "network": network_commands,
        "system": system_commands,
    }

    return docker_commands[resource][command](**kwargs)



from getpass import getpass
from os import path
from typing import TYPE_CHECKING, Type, Union

from paramiko import (
    DSSKey,
    ECDSAKey,
    Ed25519Key,
    PasswordRequiredException,
    PKey,
    RSAKey,
    SSHException,
)

import pyinfra
from pyinfra.api.exceptions import ConnectError, PyinfraError

if TYPE_CHECKING:
    from pyinfra.api.host import Host
    from pyinfra.api.state import State


def raise_connect_error(host: "Host", message, data):
    message = "{0} ({1})".format(message, data)
    raise ConnectError(message)


def _load_private_key_file(filename: str, key_filename: str, key_password: str):
    exception: Union[PyinfraError, SSHException] = PyinfraError("Invalid key: {0}".format(filename))

    key_cls: Union[Type[RSAKey], Type[DSSKey], Type[ECDSAKey], Type[Ed25519Key]]

    for key_cls in (RSAKey, DSSKey, ECDSAKey, Ed25519Key):
        try:
            return key_cls.from_private_key_file(
                filename=filename,
            )

        except PasswordRequiredException:
            if not key_password:
                # If password is not provided, but we're in CLI mode, ask for it. I'm not a
                # huge fan of having CLI specific code in here, but it doesn't really fit
                # anywhere else without duplicating lots of key related code into cli.py.
                if pyinfra.is_cli:
                    key_password = getpass(
                        "Enter password for private key: {0}: ".format(
                            key_filename,
                        ),
                    )

                # API mode and no password? We can't continue!
                else:
                    raise PyinfraError(
                        "Private key file ({0}) is encrypted, set ssh_key_password to "
                        "use this key".format(key_filename),
                    )

            try:
                return key_cls.from_private_key_file(
                    filename=filename,
                    password=key_password,
                )
            except SSHException as e:  # key does not match key_cls type
                exception = e
        except SSHException as e:  # key does not match key_cls type
            exception = e
    raise exception


def get_private_key(state: "State", key_filename: str, key_password: str) -> PKey:
    if key_filename in state.private_keys:
        return state.private_keys[key_filename]

    ssh_key_filenames = [
        # Global from executed directory
        path.expanduser(key_filename),
    ]

    if state.cwd:
        # Relative to the CWD
        path.join(state.cwd, key_filename)

    key = None
    key_file_exists = False

    for filename in ssh_key_filenames:
        if not path.isfile(filename):
            continue

        key_file_exists = True

        try:
            key = _load_private_key_file(filename, key_filename, key_password)
            break
        except SSHException:
            pass

    # No break, so no key found
    if not key:
        if not key_file_exists:
            raise PyinfraError("No such private key file: {0}".format(key_filename))
        raise PyinfraError("Invalid private key file: {0}".format(key_filename))

    # Load any certificate, names from OpenSSH:
    # https://github.com/openssh/openssh-portable/blob/049297de975b92adcc2db77e3fb7046c0e3c695d/ssh-keygen.c#L2453  # noqa: E501
    for certificate_filename in (
        "{0}-cert.pub".format(key_filename),
        "{0}.pub".format(key_filename),
    ):
        if path.isfile(certificate_filename):
            key.load_certificate(certificate_filename)

    state.private_keys[key_filename] = key
    return key



from __future__ import annotations

from pyinfra.api import operation

from . import postgres


@operation(is_idempotent=False, is_deprecated=True, deprecated_for="postgres.sql")
def sql(*args, **kwargs):
    yield from postgres.sql._inner(*args, **kwargs)


@operation(is_idempotent=False, is_deprecated=True, deprecated_for="postgres.role")
def role(*args, **kwargs):
    yield from postgres.role._inner(*args, **kwargs)


@operation(is_idempotent=False, is_deprecated=True)
def database(*args, **kwargs):
    yield from postgres.database._inner(*args, **kwargs)


@operation(is_idempotent=False, is_deprecated=True)
def dump(*args, **kwargs):
    yield from postgres.dump._inner(*args, **kwargs)


@operation(is_idempotent=False, is_deprecated=True)
def load(*args, **kwargs):
    yield from postgres.load._inner(*args, **kwargs)



import logging

import click

from pyinfra import logger, state
from pyinfra.context import ctx_state


class LogHandler(logging.Handler):
    def emit(self, record):
        try:
            message = self.format(record)
            click.echo(message, err=True)
        except Exception:
            self.handleError(record)


class LogFormatter(logging.Formatter):
    previous_was_header = True

    level_to_format = {
        logging.DEBUG: lambda s: click.style(s, "green"),
        logging.WARNING: lambda s: click.style(s, "yellow"),
        logging.ERROR: lambda s: click.style(s, "red"),
        logging.CRITICAL: lambda s: click.style(s, "red", bold=True),
    }

    def format(self, record):
        message = record.msg

        if record.args:
            message = record.msg % record.args

        # Add path/module info for debug
        if record.levelno is logging.DEBUG:
            path_start = record.pathname.rfind("pyinfra")

            if path_start:
                pyinfra_path = record.pathname[path_start:-3]  # -3 removes `.py`
                module_name = pyinfra_path.replace("/", ".")
                message = "[{0}] {1}".format(module_name, message)

        # We only handle strings here
        if isinstance(message, str):
            if ctx_state.isset() and record.levelno is logging.WARNING:
                state.increment_warning_counter()

            if "-->" in message:
                if not self.previous_was_header:
                    click.echo(err=True)
            else:
                message = "    {0}".format(message)

            if record.levelno in self.level_to_format:
                message = self.level_to_format[record.levelno](message)

            self.previous_was_header = "-->" in message
            return message

        # If not a string, pass to standard Formatter
        return super().format(record)


def setup_logging(log_level, other_log_level=None):
    if other_log_level:
        logging.basicConfig(level=other_log_level)

    logger.setLevel(log_level)
    handler = LogHandler()
    formatter = LogFormatter()
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False



from __future__ import annotations

from pyinfra.api import FactBase

from .util.packaging import parse_packages

GEM_REGEX = r"^([a-zA-Z0-9\-\+\_]+)\s\(([0-9\.]+)\)$"


class GemPackages(FactBase):
    """
    Returns a dict of installed gem packages:

    .. code:: python

        {
            'package_name': ['version'],
        }
    """

    def command(self) -> str:
        return "gem list --local"

    def requires_command(self) -> str:
        return "gem"

    default = dict

    def process(self, output):
        return parse_packages(GEM_REGEX, output)



try:
    try:
        from importlib_metadata import version
    except ImportError:
        from importlib.metadata import version

    __version__ = version("pyinfra")
except Exception:
    __version__ = "unknown"
